/*
 * ***********************************************************************
 *
 * File 'tkCanvasPlusTags.c'.
 *
 * CanvasPlus widget tag implementation.
 *
 * Written by Jorge Lima in Nov 23 14:44:03 WET 2004
 *
 */

#include <tcl.h>
#include <tk.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "tkCanvasPlusTags.h"
#include "tkCanvasPlusDrawing.h"
#include "tkCanvasPlusModes.h"
#include "tkCanvasPlus.h"

#ifdef USE_INTERNAL_HASH_TABLE
static Tcl_HashTable _TagsTable;
static int _TagsTableUsage = 0;
#endif

static Tk_Uid all_tag_uid;
static Tk_Uid current_tag_uid;


/*----------------------------------------------------------------------
 * findTagInItem --
 *	search for the given tag uid in the given item
 * Results
 *	the index, within the given item for the tag or -1 if tag was
 *      not found.
 * Side effects:
 *	None
 *---------------------------------------------------------------------*/
static int findTagInItem(CP_Item *item, Tk_Uid tag)
{
  CP_Tags *tags = &item->tags;
  int count;
  
  for (count = 0; count < tags->number; count++) {
    if ( tags->array[count] == tag )
      return count;
  }
  return -1;
}

/*----------------------------------------------------------------------
 * CPInitTags --
 *      Initializes TAGS package
 * Results:
 *	None
 * Side effects:
 *	TAGS package will be ready to use
 *---------------------------------------------------------------------*/
void CPInitTags()
{
#ifdef USE_INTERNAL_HASH_TABLE
  if (!_TagsTableUsage) {
    int isNew;
    Tcl_HashEntry *entry;
    // Init hash table
    Tcl_InitHashTable(&_TagsTable, TCL_STRING_KEYS);
    // create fast tag access for "all" and "current"
    entry = Tcl_CreateHashEntry(&_TagsTable, "all", isNew);
    all_tag_uid = (Tk_Uid) Tcl_GetHashKey(&_TagsTable, entry);
    entry = Tcl_CreateHashEntry(&_TagsTable, "current", isNew);
    current_tag_uid = (Tk_Uid) Tcl_GetHashKey(&_TagsTable, entry);
  }
  _TagsTableUsage++;
#else
  all_tag_uid = Tk_GetUid("all");
  current_tag_uid = Tk_GetUid("current");
#endif
}

/*----------------------------------------------------------------------
 * CPReleaseTags --
 *      To be called when finished with the tags package. There should
 *      exist one call to CPDeleteTags for each call to CPInitTags
 * Results:
 *	None
 * Side effects:
 *	If nobody else is using the TAGS package, memory will be released
 *---------------------------------------------------------------------*/

void CPReleaseTags()
{
#ifdef USE_INTERNAL_HASH_TABLE
  if (_TagsTableUsage==1) {
    Tcl_DeleteHashTable(&_TagsTable);
  }
  _TagsTableUsage--;
#endif
}

/*----------------------------------------------------------------------
 * CPGetTagUid --
 *      Gets the Tag Uid
 * Results:
 *	Tag Uid
 * Side effects:
 *	Eventualy a new tag is created
 *---------------------------------------------------------------------*/
Tk_Uid CPGetTagUid(const char *str)
{
#ifdef USE_INTERNAL_HASH_TABLE
  int isNew;
  Tcl_HashEntry *entry;
  entry = Tcl_CreateHashEntry(&_TagsTable, str, isNew);
  return (Tk_Uid) Tcl_GetHashKey(&_TagsTable, entry);
#else
  return Tk_GetUid(str);
#endif
}

/*----------------------------------------------------------------------
 * CPGetTagOrId --
 *      Gets tag Uid associated with tagOrId if tagOrId is a tag. if
 *      tagOrId is a number, then the pointer to the corresponding 
 *      CP_Item is returned
 * Results:
 *	Unique Tag identifier or a pointer to item depending on tagOrId
 * Side effects:
 *	None
 *---------------------------------------------------------------------*/
ClientData CPGetTagOrId(CP_Mode *mode, const char *tagOrId)
{
  ClientData object = 0;
  if (isdigit((unsigned char)tagOrId[0])) {
    char *end;
    int id = strtoul(tagOrId, &end, 0);
    if (*end != 0) goto bindByTag;
    object = (ClientData)CPGetItemFromId(mode, id);
  } else {
  bindByTag:
    object = (ClientData)CPGetTagUid(tagOrId);
  }
  return object;
}

/*----------------------------------------------------------------------
 * CPGetAllTagUid --
 *      Fast version of CPGetTagUid for tag "all"
 * Results:
 *	Tag Uid
 * Side effects:
 *	Eventualy a new tag is created
 *---------------------------------------------------------------------*/
Tk_Uid CPGetAllTagUid() { return all_tag_uid; }
Tk_Uid CPGetCurrentTagUid() { return current_tag_uid; }

/*----------------------------------------------------------------------
 * CPInitItemTags --
 *      Initializes tag struct in item
 * Results:
 *	None
 * Side effects:
 *	Item's tag struct will be ready to use
 *---------------------------------------------------------------------*/
void CPInitItemTags(CP_Item *item)
{
  CP_Tags *tags = &item->tags;
  tags->space = CP_TAGS_SPACE;
  tags->number = 0;
  tags->array = tags->staticArray;
}

void CPDeleteItemTags(CP_Item *item)
{
  CP_Tags *tags = &item->tags;
  if (tags->array != tags->staticArray)
    ckfree((char *)tags->array);
  tags->array = tags->staticArray;
}

/*----------------------------------------------------------------------
 * CPAddBindTag --
 *	append a new tag with uid to the given item
 * Results
 *	None
 * Side effects:
 *	Eventually, the item will tag storage space will be increased. 
 *---------------------------------------------------------------------*/
void CPAddBindTag(CP_Item *item, const Tk_Uid tag)
{
    CP_Tags *tags = &item->tags;

    if (findTagInItem(item, tag) >= 0)
      return;
    /*
     * Grow the tag space if there's no more room left in the current
     * block.
     */
    if (tags->number == tags->space) {
	Tk_Uid *newTagPtr;
	tags->space += 10;
	newTagPtr = (Tk_Uid *) ckalloc((unsigned)(tags->space*sizeof(Tk_Uid)));
	memcpy((void *)newTagPtr, (void *)tags->array,
	       (unsigned)(tags->number*sizeof(Tk_Uid)));
	if (tags->array != tags->staticArray) {
	  ckfree((char *)tags->array);
	}
	tags->array = newTagPtr;
    }

    // Add in the new tag.
    tags->array[tags->number] = tag;
    tags->number++;
}

/*----------------------------------------------------------------------
 * CPRemoveBindTag --
 *	remove the given tag uid from the given item
 * Results
 *	None
 * Side effects:
 *	Tag will be removed from item.
 *	Eventually, the item will tag storage space will be decreased. 
 *---------------------------------------------------------------------*/
void CPRemoveBindTag(CP_Item *item, const  Tk_Uid tag)
{
  int tagPos;
  int threshold = 2;  // threshold to avoid to much alloc/free
                      // calls when working in the boundary
  CP_Tags *tags = &item->tags;
    
  if ((tagPos = findTagInItem(item, tag)) < 0)
    return;
  
  tags->number--;
  tags->array[tagPos] = tags->array[tags->number];

  // restore tag array to static space when the quantity justifies it
  if ( (tags->array != tags->staticArray) &&
       (tags->number < CP_TAGS_SPACE - threshold) ) {
    tags->space = CP_TAGS_SPACE;
    memcpy((void *)tags->staticArray, 
	   (void *)tags->array,(tags->number*sizeof(Tk_Uid)));
    ckfree((char *)tags->array);
    tags->array = tags->staticArray;
  }
}

/*----------------------------------------------------------------------
 * CPSearchBindTag --
 *	Search for tag starting with the given item
 * Results
 *	pointer to the next tagged item
 * Side effects:
 *      None
 *---------------------------------------------------------------------*/
CP_Item * CPSearchBindTag(CP_Item *start, const  Tk_Uid tag)
{
  // This is not the best way to look for tags, besides, it
  // takes too much time with long item lists or long tag lists
  // Tag storage must be redesigned...
  while (start!=NULL) {
    if (findTagInItem(start, tag) >= 0)
      return start;
    start = start->next;
  }
  return NULL;
}

/*----------------------------------------------------------------------
 * CPGetBindScript --
 *      Sets the TCL result to the script associated with the tag and 
 *      the sequence
 * Results:
 *	TCL return value indicating error condition
 * Side effects:
 *	Tcl result string will be set
 *---------------------------------------------------------------------*/
int CPGetBindScript(Tcl_Interp *interp, Tk_BindingTable bindTbl, 
		    ClientData tagOrId, char *sequence)
{
  int result = TCL_OK;
  CONST char *command;

  // bindings are performed in the master interp
  command = Tk_GetBinding(interp, bindTbl, tagOrId, sequence);
  //  printf("BS -- %s (%p) -- %s\n", tagOrId, tagOrId, command);
  if (command == NULL) {
    Tcl_Obj *obj = Tcl_GetObjResult(interp); 
    if (Tcl_GetCharLength(obj) > 0) {
      result = TCL_ERROR;
      Tcl_SetObjResult(interp, obj);
    } else {
      Tcl_ResetResult(interp);
    }
  } else {
    Tcl_SetResult(interp, (char *)command, TCL_STATIC);
  }
  return result;
}

/*----------------------------------------------------------------------
 * CPBindScriptToTag --
 *	Register or deletes a script with a bind tag
 * Results:
 *	TCL return value indicating error condition
 * Side effects:
 *	Eventually the bindingTable will be changed
 *---------------------------------------------------------------------*/
int CPBindScriptToTag(Tcl_Interp *interp, Tk_BindingTable bindTbl,
		      ClientData tagOrId, char *sequence, char *script)
{
  int result = TCL_OK;

  if (script[0] == '\0') {
    result = Tk_DeleteBinding(interp, bindTbl, tagOrId, sequence);
  }
  else {
    unsigned long mask;
    int append = 0;
    if (script[0] == '+') {
      append = 1;
      script++;
    }
    mask = Tk_CreateBinding(interp, bindTbl, tagOrId, 
			    sequence, script, append);

    if (mask == 0) {
      Tcl_AppendResult(interp, "mask is 0", (char *) NULL);
      result = TCL_ERROR;
    }
    else if (mask & (unsigned) ~(ButtonMotionMask|Button1MotionMask
				 |Button2MotionMask|Button3MotionMask
				 |Button4MotionMask|Button5MotionMask
				 |ButtonPressMask|ButtonReleaseMask
				 |EnterWindowMask|LeaveWindowMask
				 |KeyPressMask|KeyReleaseMask
				 |PointerMotionMask|VirtualEventMask)) {
      Tk_DeleteBinding(interp, bindTbl, tagOrId, sequence);
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, "requested illegal events; ",
		       "only key, button, motion, enter, leave, and virtual ",
		       "events may be used", (char *) NULL);
      result = TCL_ERROR;
    }
    else {
      printf("binding of %s %s {%s}\n", (char *)tagOrId, sequence, script);
    }
  }
  return result;
}

